from unittest import TestCase
from utils import load_deployment, relations_json_to_tuples


class TestLoadDeployment(TestCase):

    def test_inherits_one_level(self):
        """
        One level of inheritance is supported just fine.
        """
        config = {
            'foo': {
                'series': 'precise',
                'services': {
                    'foo-1': {
                        'options': {
                            'foo': True,
                            },
                        },
                    'foo-2': {},
                    },
                },
            'bar': {
                'series': 'lucid',
                'inherits': 'foo',
                'services': {
                    'foo-1': {
                        'options': {
                            'foo': False,
                            'bar': True,
                            },
                        },
                    'bar-1': {}
                    },
                }
            }

        (series, charms,
         relations, overrides) = load_deployment(config, 'bar')
        self.assertEqual('lucid', series)
        self.assertEqual({
            'foo-1': {
                'options': {
                    'foo': False,
                    'bar': True,
                    },
                },
            'foo-2': {},
            'bar-1': {},
            }, charms)
        self.assertEqual({}, relations)
        self.assertEqual({}, overrides)

    def test_inherits_two_levels(self):
        """
        Muliple recursive levels of inheritance are also supported.
        """
        config = {
            'foo': {
                'series': 'precise',
                'services': {
                    'foo-1': {
                        'options': {
                            'foo': True,
                            },
                        },
                    'foo-2': {},
                    },
                },
            'bar': {
                'series': 'lucid',
                'inherits': 'foo',
                'services': {
                    'foo-1': {
                        'options': {
                            'foo': False,
                            'bar': True,
                            },
                        },
                    },
                },
            'baz': {
                'series': 'precise',
                'inherits': 'bar',
                'services': {
                    'foo-1': {
                        'options': {
                            'baz': True,
                            },
                        },
                    'baz-1': {},
                    },
                },
            }

        (series, charms,
         relations, overrides) = load_deployment(config, 'baz')
        self.assertEqual('precise', series)
        self.assertEqual({
            'foo-1': {
                'options': {
                    'foo': False,
                    'bar': True,
                    'baz': True,
                    },
                },
            'foo-2': {},
            'baz-1': {},
            }, charms)
        self.assertEqual({}, relations)
        self.assertEqual({}, overrides)

    def test_inherits_multiple(self):
        """
        It is also possible to inherit from multiple deployments at a single
        level, which may or may not recurse.
        """
        config = {
            'foo': {
                'series': 'precise',
                'services': {
                    'foo-1': {
                        'options': {
                            'foo': True,
                            },
                        },
                    'foo-2': {},
                    },
                },
            'bar': {
                'series': 'lucid',
                'services': {
                    'foo-1': {
                        'options': {
                            'foo': False,
                            'bar': True,
                            },
                        },
                    'bar-1': {},
                    },
                },
            'baz': {
                'series': 'quantal',
                'inherits': ['foo', 'bar'],
                'services': {
                    'foo-1': {
                        'options': {
                            'baz': True,
                            },
                        },
                    'baz-1': {},
                    },
                },
            }

        (series, charms,
         relations, overrides) = load_deployment(config, 'baz')
        self.assertEqual('quantal', series)
        self.assertEqual({
            'foo-1': {
                'options': {
                    'foo': False,
                    'bar': True,
                    'baz': True,
                    },
                },
            'foo-2': {},
            'bar-1': {},
            'baz-1': {},
            }, charms)
        self.assertEqual({}, relations)
        self.assertEqual({}, overrides)

    def test_inherits_multiple_recurse(self):
        """
        It is also possible to inherit from multiple deployments at a single
        level, one of which inherits from another deployment.
        """
        config = {
            'qux': {
                'series': 'precise',
                'services': {
                    'foo-1': {
                        'options': {
                            'qux': True,
                            },
                        },
                    'foo-2': {},
                    },
                },
            'foo': {
                'inherits': 'qux',
                'series': 'precise',
                'services': {
                    'foo-1': {
                        'options': {
                            'foo': True,
                            },
                        },
                    'foo-2': {},
                    },
                },
            'bar': {
                'series': 'lucid',
                'services': {
                    'foo-1': {
                        'options': {
                            'foo': False,
                            'bar': True,
                            },
                        },
                    'bar-1': {},
                    },
                },
            'baz': {
                'series': 'quantal',
                'inherits': ['foo', 'bar'],
                'services': {
                    'foo-1': {
                        'options': {
                            'baz': True,
                            },
                        },
                    'baz-1': {},
                    },
                },
            }

        (series, charms,
         relations, overrides) = load_deployment(config, 'baz')
        self.assertEqual('quantal', series)
        self.assertEqual({
            'foo-1': {
                'options': {
                    'qux': True,
                    'foo': False,
                    'bar': True,
                    'baz': True,
                    },
                },
            'foo-2': {},
            'bar-1': {},
            'baz-1': {},
            }, charms)
        self.assertEqual({}, relations)
        self.assertEqual({}, overrides)


class TestRelationsJSONToTuples(TestCase):

    def test_key_by_weight(self):
        """
        Relations are grouped by weight.
        """
        relations = {
            'foo': {
                'weight': 42,
                'consumes': ['bar'],
                },
            'baz': {
                'weight': 60,
                'consumes': ['qux'],
                }
            }
        self.assertEqual({42: [('foo', 'bar')],
                          60: [('baz', 'qux')]},
                         relations_json_to_tuples(relations))

    def test_same_weight_overrides_deterministic(self):
        """
        If a weight is specified and two relations have the same weight, the
        last one in alphabetical consumer order wins.
        """
        relations = {
            'foo': {
                'weight': 42,
                'consumes': ['bar'],
                },
            'baz': {
                'weight': 42,
                'consumes': ['qux'],
                }
            }
        self.assertEqual({42: [('foo', 'bar')]},
                         relations_json_to_tuples(relations))

    def test_weight_optional(self):
        """
        If a weight is not provided, all relations are taken into account with
        no overrides.
        """
        relations = {
            'foo': {
                'consumes': ['bar'],
                },
            'baz': {
                'consumes': ['qux'],
                }
            }
        self.assertEqual({0: [('baz', 'qux'), ('foo', 'bar')]},
                         relations_json_to_tuples(relations))
